"""Module containing the handler for the Quantum Inspire persistent configuration."""

from __future__ import annotations

import asyncio
import json
import time
from os import PathLike
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Type, cast

import jwt
import typer
from compute_api_client import ApiClient, AuthConfigApi, Configuration, MembersApi
from pydantic import BaseModel, BeforeValidator, HttpUrl
from pydantic.fields import Field, FieldInfo
from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, SettingsConfigDict
from typing_extensions import Annotated

Url = Annotated[str, BeforeValidator(lambda value: str(HttpUrl(value)).rstrip("/"))]


def ensure_config_file_exists(file_path: Path, file_encoding: Optional[str] = None) -> None:
    """Create the file if it does not exist.

    Args:
        file_path: the file path.
        file_encoding: The encoding of the file.
    """
    if not file_path.exists():
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.open("w", encoding=file_encoding).write('{"auths": {}}')


class JsonConfigSettingsSource(PydanticBaseSettingsSource):
    """Settings source that loads variables from a JSON file specified in the Config class.

    Returns:
        A dictionary with the setting variables from the JSON file.
    """

    def get_field_value(self, field: FieldInfo, field_name: str) -> Tuple[Any, str, bool]:
        raise NotImplementedError

    def __call__(self) -> Any:
        encoding = self.config.get("env_file_encoding")

        assert isinstance(self.config["json_file"], PathLike)

        json_config_file = Path(self.config["json_file"])

        ensure_config_file_exists(json_config_file, encoding)
        return json.loads(json_config_file.read_text(encoding))


class TokenInfo(BaseModel):
    """A pydantic model for storing all information regarding oauth access and refresh tokens."""

    access_token: str
    expires_in: int
    refresh_token: str
    generated_at: float = Field(default_factory=time.time)

    @property
    def access_expires_at(self) -> float:
        """Timestamp containing the time when the access token will expire."""
        return self.generated_at + self.expires_in


class AuthSettings(BaseModel):
    """Pydantic model for storing all auth related settings for a given host."""

    client_id: str = "compute-job-manager"
    audience: str = "compute-job-manager"
    # Keycloak requires api-access in scope for compute-job-manager audience
    # Auth0 requires offline_access in scopefor sending a refresh token
    scope: str = "api-access openid profile email offline_access"
    code_challenge_method: str = "S256"
    code_verifyer_length: int = 64
    well_known_endpoint: Url = (
        "https://auth.qi2.quantum-inspire.com/realms/oidc_production/.well-known/openid-configuration"
    )
    tokens: Optional[TokenInfo] = None
    team_member_id: Optional[int] = None

    @property
    def owner_id(self) -> int:
        if self.team_member_id is None:
            raise ValueError("Please set the default team_member_id for this host!")
        return self.team_member_id


class Settings(BaseSettings):  # pylint: disable=too-few-public-methods
    """The settings class for the Quantum Inspire persistent configuration."""

    model_config = SettingsConfigDict(
        env_file_encoding="utf-8",
        env_prefix="QI2_",
        json_file=Path.joinpath(Path.home(), ".quantuminspire", "config.json"),
    )

    auths: Dict[Url, AuthSettings]

    default_host: Url = "https://api.quantum-inspire.com"

    @property
    def default_auth_settings(self) -> AuthSettings:
        return self.auths[self.default_host]

    # R0913: Too many arguments (6/5) (too-many-arguments)
    @classmethod
    def settings_customise_sources(  # pylint: disable=R0913
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Customise the settings sources (by adding, removing and changing the order of sources).

        Args:
            init_settings: The initial settings (Settings object creation): highest priority.
            env_settings: The configuration settings (Config inner object creation).
            file_secret_settings: The file secret settings: lowest priority

        Returns:
            The original sources, with
            - the JSON file as source added after the env settings and before the file secret settings.
            The order determines the priority!
        """
        return (
            init_settings,
            env_settings,
            JsonConfigSettingsSource(settings_cls),
            file_secret_settings,
        )

    async def fetch_auth_settings(self, host: Optional[Url] = None) -> None:
        """Fetch suggested auth settings for the default host."""
        if host is None:
            host = self.default_host
        async with ApiClient(Configuration(host=host)) as api_client:
            auth_config = await AuthConfigApi(api_client).auth_config_auth_config_get()
            self.auths[host] = AuthSettings(
                client_id=auth_config.client_id,
                audience=auth_config.audience,
                well_known_endpoint=auth_config.well_known_endpoint,
            )

    def store_tokens(self, host: Url, tokens: TokenInfo) -> None:
        """

        :param host: the hostname of the api for which the tokens are intended
        :param tokens: OAuth access and refresh tokens
        :return: None

        This functions stores the team_member_id, access and refresh tokens in the config.json file.
        """
        self.auths[host].tokens = tokens
        member_id = self.get_team_member_id(host=host, access_token=tokens.access_token)
        self.auths[host].team_member_id = member_id
        self.write_settings_to_file()

    def write_settings_to_file(self) -> None:
        assert isinstance(self.model_config["json_file"], PathLike)
        Path(self.model_config["json_file"]).write_text(
            self.model_dump_json(indent=2), encoding=self.model_config.get("env_file_encoding")
        )

    @staticmethod
    async def _fetch_team_member_id(host: str, access_token: str) -> int:
        config = Configuration(host=host, access_token=access_token)
        async with ApiClient(config) as api_client:
            api_instance = MembersApi(api_client)
            members_page = await api_instance.read_members_members_get()
            members = members_page.items
            if len(members) == 1:
                member_id = members[0].id
                return cast(int, member_id)

            typer.echo("Choose a member ID from the list for project configuration.")
            json_string = "[" + ",".join(member.model_dump_json(indent=4) for member in members) + "]"
            typer.echo(json_string)

            member_ids = [member.id for member in members]

            while True:
                try:
                    member_id = int(input("Please enter one of the given ids: "))
                    if member_id not in member_ids:
                        raise ValueError
                    return cast(int, member_id)
                except ValueError:
                    typer.echo("Invalid input. Please enter a valid id or CTRL + C to cancel.")

    @classmethod
    def get_team_member_id(cls, host: str, access_token: str) -> int:
        return asyncio.run(cls._validate_token_and_retrieve_team_member_id(host, access_token))

    @classmethod
    async def _validate_token_and_retrieve_team_member_id(cls, host: str, access_token: str) -> int:
        await cls._wait_until_token_becomes_valid(access_token)
        return await cls._fetch_team_member_id(host, access_token)

    @staticmethod
    async def _wait_until_token_becomes_valid(access_token: str) -> None:
        decoded_token = jwt.decode(access_token, options={"verify_signature": False})
        token_issued_at = int(decoded_token["iat"])

        while True:
            current_time = int(time.time())
            time_diff = token_issued_at - current_time

            if time_diff > 0:
                await asyncio.sleep(time_diff)
            return



"""Module containing the commands for the Quantum Inspire CLI."""

import asyncio
import webbrowser
from pathlib import Path
from typing import Optional

import typer
from compute_api_client import JobStatus
from rich import print
from rich.console import Console
from rich.table import Table
from typer import Typer

from quantuminspire.sdk.models.cqasm_algorithm import CqasmAlgorithm
from quantuminspire.sdk.models.file_algorithm import FileAlgorithm
from quantuminspire.sdk.models.hybrid_algorithm import HybridAlgorithm
from quantuminspire.sdk.models.job_options import JobOptions
from quantuminspire.util.api.local_backend import LocalBackend
from quantuminspire.util.api.remote_backend import RemoteBackend
from quantuminspire.util.authentication import OauthDeviceSession
from quantuminspire.util.configuration import Settings, Url

app = Typer(add_completion=False, no_args_is_help=True)
backend_types_app = Typer(no_args_is_help=True)
app.add_typer(backend_types_app, name="backends", help="Manage backends")
files_app = Typer(no_args_is_help=True)
app.add_typer(files_app, name="files", help="Manage files")
results_app = Typer(no_args_is_help=True)
app.add_typer(results_app, name="results", help="Manage results")
final_results_app = Typer(no_args_is_help=True)
app.add_typer(final_results_app, name="final_results", help="Manage final results")

console = Console()


@backend_types_app.command("list")
def list_backend_types() -> None:
    """List backends.

    List all available backends and their configuration.
    """
    backend = RemoteBackend()
    backend_types = backend.get_backend_types().items
    table = Table("id", "name", "status", "is_hardware", "supports_raw_data", "number_of_qubits", "max_number_of_shots")
    for backend_type in backend_types:
        table.add_row(
            str(backend_type.id),
            backend_type.name,
            str(backend_type.status.name),
            str(backend_type.is_hardware),
            str(backend_type.supports_raw_data),
            str(backend_type.nqubits),
            str(backend_type.max_number_of_shots),
        )
    console.print(table)


@backend_types_app.command("get")
def get_backend_types(backend_type_id: int = typer.Argument(..., help="The id of the backend type")) -> None:
    """Get a backend.

    Get the configuration of a specific backend.
    """
    backend = RemoteBackend()
    backend_types = backend.get_backend_types().items

    backend_type = next((bt for bt in backend_types if bt.id == backend_type_id), None)
    if backend_type is None:
        console.print(f"Backend type with id {backend_type_id} not found.")
        raise typer.Exit(1)
    else:
        console.print(backend_type.model_dump())


def load_algorithm_from_file(file_path: Path) -> FileAlgorithm:
    """Load an algorithm from a file."""
    if file_path.suffix == ".py":
        algorithm: FileAlgorithm = HybridAlgorithm("", str(file_path))
    elif file_path.suffix == ".cq":
        algorithm = CqasmAlgorithm("", str(file_path))
    else:
        raise ValueError(f"Unsupported file type: {file_path.suffix}. Supported types are .py and .cq")

    algorithm.read_file(file_path)
    return algorithm


@files_app.command("upload")
def upload_files(
    name: str = typer.Argument(..., help="The name of the file to upload"),
    backend_type_id: int = typer.Argument(
        ..., help="The id of the backend type on which this algorithm should be executed"
    ),
    num_shots: int = typer.Option(
        1024, help="The number of shots to run the algorithm (only for pure cQASM algorithms)"
    ),
    store_raw_data: bool = typer.Option(
        False, help="Whether to enable shot memory for the algorithm (only for pure cQASM algorithms)"
    ),
) -> None:
    """Upload a file to the QI API.

    Upload a file containing either a Hybrid (.py) or a Quantum (.cq) algorithm, and run it on the QI platform.
    """
    backend = RemoteBackend()
    program = load_algorithm_from_file(Path(name))
    program.read_file(Path(name))
    job_options = JobOptions(number_of_shots=num_shots, raw_data_enabled=store_raw_data)
    job_id = backend.run(program, backend_type_id=backend_type_id, options=job_options)
    typer.echo(f"Upload file with name: {name}")
    typer.echo(f"job_id {job_id}")


@files_app.command("run")
def run_file(
    name: str = typer.Argument(..., help="The full path of the file to run"),
) -> None:
    """Run a file locally.

    Run a Hybrid Quantum/Classical Algorithm locally. The quantum part will be run with QXEmulator.
    """
    algorithm = HybridAlgorithm("test", "test")
    algorithm.read_file(Path(name))

    local_backend = LocalBackend()
    job_id = local_backend.run(algorithm, 0)
    results = local_backend.get_results(job_id)
    typer.echo(f"{results}")


def _has_job_failed(backend: RemoteBackend, job_id: int) -> None:
    """Check if a job has failed and exit with error_code 1 if it has."""
    job = backend.get_job(job_id)
    if job.status == JobStatus.FAILED:
        job.message = job.message or "Job failed."
        typer.echo(job.message, err=True)
        typer.echo(f"Trace id: {job.trace_id}", err=True)
        raise typer.Exit(1)


@results_app.command("get")
def get_results(job_id: int = typer.Argument(..., help="The id of the run")) -> None:
    """Retrieve the results for a run.

    Takes the id as returned by upload_files and retrieves the results for that run, if it's finished.
    """
    backend = RemoteBackend()
    _has_job_failed(backend, job_id)
    results = backend.get_results(job_id)

    if results is None:
        typer.echo("No results.")
        raise typer.Exit(1)

    typer.echo("Raw results:")
    print(results.model_dump())


@final_results_app.command("get")
def get_final_results(job_id: int = typer.Argument(..., help="The id of the run")) -> None:
    """Retrieve the final results for a run.

    Takes the id as returned by upload_files and retrieves the final results for that job, if it's finished.
    """
    backend = RemoteBackend()
    _has_job_failed(backend, job_id)
    results = backend.get_final_results(job_id)

    if results is None:
        typer.echo("No final results.")
        raise typer.Exit(1)

    typer.echo("Raw final results:")
    print(results.model_dump())


@app.command("login")
def login(
    host: Optional[str] = typer.Argument(None, help="The URL of the platform to which to connect"),
    override_auth_config: bool = typer.Option(
        False,
        help="Will ignore authentication configuration suggested by the API and use stored configuration instead",
    ),
) -> None:
    """Log in to Quantum Inspire.

    Log in to the Quantum Inspire platform. The host can be overwritten, so that the user can connect to different
    instances. If no host is specified, the production environment will be used.
    """
    settings = Settings()
    host = host or settings.default_host
    host_url = Url(host)
    settings.default_host = host_url

    if not override_auth_config:
        asyncio.run(settings.fetch_auth_settings(host_url))

    auth_session = OauthDeviceSession(settings.auths[host_url])

    login_info = auth_session.initialize_authorization()
    typer.echo(f"Please continue logging in by opening: {login_info['verification_uri_complete']} in your browser")
    typer.echo(f"If promped to verify a code, please confirm it is as follows: {login_info['user_code']}")
    webbrowser.open(login_info["verification_uri_complete"], new=2)
    tokens = auth_session.poll_for_tokens()
    settings.store_tokens(host_url, tokens)
    typer.echo("Login successful!")
    typer.echo(f"Using member ID {settings.auths[host].team_member_id}")


@app.command("set-default-host")
def set_default_host(
    host: str = typer.Argument(help="The URL of the platform to which to connect"),
) -> None:
    """Set default_host for interacting with Quantum Inspire."""
    settings = Settings()
    settings.default_host = Url(host)
    settings.write_settings_to_file()
    typer.echo(f"Default host set to {host}")



"""Module containing the Quantum Circuit class."""

from types import TracebackType
from typing import Optional, Type

from compute_api_client import AlgorithmType, CompileStage
from opensquirrel import CircuitBuilder

from quantuminspire.sdk.models.base_algorithm import BaseAlgorithm


class Circuit(BaseAlgorithm):
    """A container object, interacting with OpenSquirrel and storing cQASM internally.

    A circuit wraps OpenSquirrel to handle the boilerplate code for the CircuitBuilder.
    """

    def __init__(
        self, platform_name: str, program_name: str, number_of_qubits: int, bit_registers: Optional[int] = None
    ) -> None:
        super().__init__(platform_name, program_name)
        self._number_of_qubits = number_of_qubits
        if bit_registers is None:
            bit_registers = number_of_qubits
        self._number_of_bit_registers = bit_registers
        self.ir = CircuitBuilder(qubit_register_size=number_of_qubits, bit_register_size=bit_registers)
        self._cqasm: str = ""

    @property
    def content(self) -> str:
        return self._cqasm

    @property
    def content_type(self) -> AlgorithmType:
        return AlgorithmType.QUANTUM

    @property
    def compile_stage(self) -> CompileStage:
        return CompileStage.NONE

    def __enter__(self) -> "Circuit":
        self.initialize()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> bool:
        self.finalize()
        return exc_type is None

    @property
    def language_name(self) -> str:
        return "cQASM"

    def initialize(self) -> None:
        """Initialize the quantum circuit."""

    def finalize(self) -> None:
        """Finalize the quantum circuit.

        After finishing writing the quantum circuit various actions are performed to generate the actual cQASM circuit.
        The circuit is converted to cQASM. Due to some differences between OpenSquirrel and QX, the casing of the gates
        need to be tweaked.
        """
        self._cqasm = str(self.ir.to_circuit())
        self._cqasm = (
            self._cqasm.replace("h", "H")
            .replace("cnot", "CNOT")
            .replace("x", "X")
            .replace("y", "Y")
            .replace("z", "Z")
            .replace("x90", "X90")
            .replace("mx90", "mX90")
            .replace("y90", "Y90")
            .replace("my90", "mY90")
            .replace("RZ", "Rz")
            .replace("RX", "Rx")
            .replace("cz", "CZ")
        )



"""Quantum Inspire SDK.

Copyright 2022 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from abc import ABC, abstractmethod
from typing import Any

from quantuminspire.sdk.models.base_algorithm import BaseAlgorithm


class BaseBackend(ABC):
    """Base interface for backends.

    A backend will execute the hybrid algorithm, or quantum circuit provided.
    """

    @abstractmethod
    def run(self, program: BaseAlgorithm, backend_type_id: int) -> int:
        """Execute provided algorithm/circuit."""
        raise NotImplementedError

    @abstractmethod
    def get_results(self, job_id: int) -> Any:
        """Get results for algorithm/circuit."""
        raise NotImplementedError



"""Quantum Inspire SDK.

Copyright 2022 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from typing import TYPE_CHECKING, Any, List

from quantuminspire.util.api.quantum_interface import ExecuteCircuitResult, QuantumInterface as QuantumInterfaceProtocol

if TYPE_CHECKING:  # pragma: no cover
    from quantuminspire.util.api.local_backend import LocalBackend


class QuantumInterface(QuantumInterfaceProtocol):
    """Quantum Interface implementation for running quantum circuits from hybrid algorithms."""

    # pylint: disable = R0903
    # R0903: Too few public methods (1/2) (too-few-public-methods)

    def __init__(self, backend: "LocalBackend"):
        self.backend = backend
        self.results: List[Any] = []

    def execute_circuit(self, circuit: str, number_of_shots: int) -> ExecuteCircuitResult:
        """Execute a quantum circuit."""
        results = self.backend.run_quantum(circuit, number_of_shots)
        self.results.append(results.results)
        return results



"""Quantum Inspire SDK.

Copyright 2022 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from pydantic import BaseModel


class ExecuteCircuitResult(BaseModel):
    """Result of executing a quantum circuit."""

    results: Dict[str, float]
    shots_requested: int
    shots_done: int


class QuantumInterface(ABC):
    """Interface for running quantum circuits from hybrid algorithms."""

    # pylint: disable = R0903
    # Too few public methods (1/2) (too-few-public-methods)

    results: List[Any]

    @abstractmethod
    def execute_circuit(self, circuit: str, number_of_shots: int) -> ExecuteCircuitResult:
        """Execute a quantum circuit."""
        raise NotImplementedError



"""Module containing the Hybrid Quantum/Classical Algorithm class."""

from compute_api_client import AlgorithmType

from quantuminspire.sdk.models.file_algorithm import FileAlgorithm


class HybridAlgorithm(FileAlgorithm):
    """A container object, reading the python algorithm and keeping metadata.

    The HybridAlgorithm reads the python file describing the algorithm and stores it in `.content`.
    """

    @property
    def content_type(self) -> AlgorithmType:
        return AlgorithmType.HYBRID

    @property
    def language_name(self) -> str:
        return "Python"



"""Module containing the Hybrid Quantum/Classical Algorithm class."""

from pathlib import Path

from compute_api_client import CompileStage

from quantuminspire.sdk.models.base_algorithm import BaseAlgorithm


class FileAlgorithm(BaseAlgorithm):
    """A container object, reading the python algorithm and keeping metadata.

    The FileAlgorithm reads the python file describing the algorithm and stores it in `.content`.
    """

    def __init__(self, platform_name: str, program_name: str) -> None:
        super().__init__(platform_name, program_name)
        self._content = ""

    @property
    def content(self) -> str:
        return self._content

    @property
    def compile_stage(self) -> CompileStage:
        return CompileStage.NONE

    def read_file(self, file_pointer: Path) -> None:
        """Read the python file to the wrapper.

        Args:
            file_pointer: the file path to the file to read.
        """
        self._content = file_pointer.read_text()



"""Quantum Inspire SDK.

Copyright 2022 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

import importlib
import sys
import types
from enum import IntEnum
from typing import Any, Union

from quantuminspire.sdk.models.base_algorithm import BaseAlgorithm
from quantuminspire.sdk.models.circuit import Circuit
from quantuminspire.sdk.models.hybrid_algorithm import HybridAlgorithm
from quantuminspire.sdk.quantum_interface import QuantumInterface
from quantuminspire.util.api.base_backend import BaseBackend
from quantuminspire.util.api.quantum_interface import ExecuteCircuitResult, QuantumInterface as QuantumInterfaceProtocol


def import_qxelarator() -> types.ModuleType:  # pragma: no cover
    """Helper function for importing optional dependency qxelarator."""

    try:
        return importlib.import_module("qxelarator")
    except ModuleNotFoundError as exc:
        print(
            "Error: Dependencies for local execution not installed, install with 'pip install quantuminspire[local]'",
            file=sys.stderr,
        )
        raise SystemExit from exc


class LocalBackend(BaseBackend):
    """Connection to remote backend.

    Create a connection with the remote backend for Quantum Inspire. This connection creates the appropriate projects,
    algorithms, files etc. The algorithm/circuit will also be run.
    """

    class JobFakeID(IntEnum):
        """Helper to conform to BaseBackend interface of returning job IDs."""

        QUANTUM = -1
        HYBRID = -2

    def __init__(self, qxelarator: Union[types.ModuleType, None] = None) -> None:
        super().__init__()
        self._quantum_results: Union[ExecuteCircuitResult, None] = None
        self._hybrid_results: Union[Any, None] = None
        self._qxelarator = qxelarator if qxelarator else import_qxelarator()

    def run(self, program: BaseAlgorithm, backend_type_id: int) -> int:
        """Execute provided algorithm/circuit."""
        if isinstance(program, HybridAlgorithm):
            quantum_interface = QuantumInterface(self)
            self._hybrid_results = self.run_hybrid(program, quantum_interface)
            return self.JobFakeID.HYBRID
        if isinstance(program, Circuit):
            self._quantum_results = self.run_quantum(program.content)
            return self.JobFakeID.QUANTUM
        raise AssertionError("Unknown algorithm type")

    def run_hybrid(self, algorithm: HybridAlgorithm, quantum_interface: QuantumInterfaceProtocol) -> Any:
        """Execute provided algorithm."""

        # pylint: disable = E1101, W0122
        # E1101: Instance of 'module' has no 'execute' member (no-member)
        # W0122: Use of exec (exec-used)

        program = types.ModuleType("mod")

        exec(algorithm.content, program.__dict__)

        if not hasattr(program, "execute") or not callable(program.execute):
            raise AssertionError("'execute' hook not found or not callable.")
        program.execute(quantum_interface)

        if not hasattr(program, "finalize") or not callable(program.finalize):
            raise AssertionError("'finalize' hook not found or not callable.")
        result = program.finalize(quantum_interface.results)
        if not isinstance(result, dict):
            raise AssertionError("'finalize' hook returned invalid type, must be dict.")
        return result

    def run_quantum(self, circuit: str, number_of_shots: int = 1) -> ExecuteCircuitResult:
        """Execute provided circuit."""
        result = self._qxelarator.execute_string(circuit, iterations=number_of_shots)
        return ExecuteCircuitResult(
            results=result.results,
            shots_done=result.shots_done,
            shots_requested=result.shots_requested,
        )

    def get_results(self, job_id: int) -> Any:
        """Get results for algorithm/circuit."""
        if job_id == self.JobFakeID.QUANTUM:
            return self._quantum_results
        if job_id == self.JobFakeID.HYBRID:
            return self._hybrid_results

        raise AssertionError("Unknown job id")



"""Module containing the Abstract Base Class for Algorithm classes."""

from abc import ABC, abstractmethod


class BaseAlgorithm(ABC):
    """A container object, used to store content and metadata for the algorithm."""

    def __init__(self, platform_name: str, program_name: str) -> None:
        self._platform_name = platform_name
        self._program_name = program_name

    @property
    def program_name(self) -> str:
        """Return the name of the program.

        Returns:
            The string representation of the program name.
        """
        return self._program_name

    @property
    def platform_name(self) -> str:
        """Return the name of the platform the algorithm is intended to run on.

        Returns:
            The string representation of the platform name.
        """
        return self._platform_name

    @property
    @abstractmethod
    def content(self) -> str:
        """Return the content.

        Returns:
            The string representation of the program.
        """

    @property
    @abstractmethod
    def content_type(self) -> str:
        """Return the content type.

        Returns:
            The string representation of the `AlgorithmType`.
        """

    @property
    @abstractmethod
    def compile_stage(self) -> str:
        """Return the compile stage.

        Returns:
            The string representation of the `CompileStage`.
        """

    @property
    @abstractmethod
    def language_name(self) -> str:
        """The name of the language the algorithm in the file is written in.

        Should match a language in the API.
        """



from pydantic import BaseModel, Field


class JobOptions(BaseModel):
    """Options for running a job.

    Subset of fields in the JobIn schema.
    """

    number_of_shots: int = Field(
        1024, ge=1, description="Number of shots for the job (only pure cQASM algorithms supported)."
    )
    raw_data_enabled: bool = Field(
        False, description="Whether to enable shot memory for the job (only pure cQASM algorithms supported)."
    )



"""Module containing the Quantum Algorithm class."""

from compute_api_client import AlgorithmType

from quantuminspire.sdk.models.file_algorithm import FileAlgorithm


class CqasmAlgorithm(FileAlgorithm):
    """A container object, reading the python algorithm and keeping metadata.

    The CqasmAlgorithm reads the python file describing the algorithm and stores it in `.content`.
    """

    @property
    def content_type(self) -> AlgorithmType:
        return AlgorithmType.QUANTUM

    @property
    def language_name(self) -> str:
        return "cQASM"



import time
from typing import Any, Dict, Tuple, cast

import compute_api_client
import requests
from oauthlib.oauth2 import Client

from quantuminspire.util.configuration import AuthSettings, TokenInfo


class AuthorisationError(Exception):
    """Indicates that the authorisation permanently went wrong."""

    pass


class AuthorisationPending(Exception):
    """Indicates that we should continue polling."""


class OauthDeviceSession:
    def __init__(self, settings: AuthSettings):
        self._settings = settings
        self._client_id = settings.client_id
        self._token_info = settings.tokens
        self._token_endpoint, self._device_endpoint = self._get_endpoints()
        self._oauth_client = Client(settings.client_id)
        self._headers = {"Content-Type": "application/x-www-form-urlencoded"}
        self.expires_in = 600  # expiration time in seconds
        self.polling_interval: float = 5  # polling interval in seconds
        self.expires_at = time.time()
        self._device_code = ""
        self._refresh_time_reduction = 5  # the number of seconds to refresh the expiration time

    def _get_endpoints(self) -> Tuple[str, str]:
        response = requests.get(self._settings.well_known_endpoint)
        response.raise_for_status()
        config = response.json()
        return config["token_endpoint"], config["device_authorization_endpoint"]

    def initialize_authorization(self) -> Dict[str, Any]:
        code_verifier = self._oauth_client.create_code_verifier(self._settings.code_verifyer_length)
        self._oauth_client.create_code_challenge(code_verifier, self._settings.code_challenge_method)
        data = {
            "client_id": self._client_id,
            "code_challenge_method": self._settings.code_challenge_method,
            "code_challenge": self._oauth_client.code_challenge,
            "audience": self._settings.audience,
            "scope": self._settings.scope,
        }

        response = requests.post(self._device_endpoint, data=data, headers=self._headers).json()
        if "error" in response:
            raise AuthorisationError(response["error_description"])

        self.expires_in = int(response["expires_in"])
        self.polling_interval = response["interval"]
        self.expires_at = time.monotonic() + self.expires_in
        self._device_code = str(response["device_code"])

        return cast(Dict[str, Any], response)

    def request_token(self) -> TokenInfo:
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "client_id": self._client_id,
            "code_verifier": self._oauth_client.code_verifier,
            "device_code": self._device_code,
        }

        response = requests.post(self._token_endpoint, data=data, headers=self._headers)

        if response.status_code >= 400:
            content = response.json()
            if content["error"] == "authorization_pending":
                raise AuthorisationPending(content["error"])
            if content["error"] == "slow_down":
                self.polling_interval += 5
                raise AuthorisationPending(content["error"])

        if response.status_code == 200:
            content = response.json()
            self._token_info = TokenInfo(**content)
            return self._token_info

        raise AuthorisationError(f"Received status code: {response.status_code}\n {response.text}")

    def poll_for_tokens(self) -> TokenInfo:
        while time.monotonic() < self.expires_at:
            try:
                return self.request_token()
            except AuthorisationPending:
                time.sleep(self.polling_interval)

        raise AuthorisationError("Login session timed out, please login again.")

    def refresh(self) -> TokenInfo:
        if self._token_info is None:
            raise AuthorisationError("You should authenticate first before you can refresh")

        if self._token_info.access_expires_at > time.time() + self._refresh_time_reduction:
            return self._token_info

        data = {
            "grant_type": "refresh_token",
            "client_id": self._client_id,
            "refresh_token": self._token_info.refresh_token,
        }

        response = requests.post(self._token_endpoint, data=data, headers=self._headers)

        if response.status_code == 200:
            self._token_info = TokenInfo(**response.json())
            return self._token_info

        raise AuthorisationError(f"Received status code: {response.status_code}\n {response.text}")


class Configuration(compute_api_client.Configuration):  # type: ignore[misc]
    def __init__(self, host: str, oauth_session: OauthDeviceSession, **kwargs: Any):
        self._oauth_session = oauth_session
        super().__init__(host=host, **kwargs)

    def auth_settings(self) -> Any:
        token_info = self._oauth_session.refresh()
        self.access_token = token_info.access_token
        return super().auth_settings()



"""Quantum Inspire SDK.

Copyright 2022 QuTech Delft

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
License. You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific
language governing permissions and limitations under the License.
"""

import asyncio
from typing import Any, List

from compute_api_client import (
    Algorithm,
    AlgorithmIn,
    AlgorithmsApi,
    ApiClient,
    BackendTypesApi,
    BatchJob,
    BatchJobIn,
    BatchJobsApi,
    Commit,
    CommitIn,
    CommitsApi,
    File,
    FileIn,
    FilesApi,
    FinalResult,
    FinalResultsApi,
    Job,
    JobIn,
    JobsApi,
    JobStatus,
    Language,
    LanguagesApi,
    Project,
    ProjectIn,
    ProjectsApi,
    Result,
    ResultsApi,
    ShareType,
)

from quantuminspire.sdk.models.base_algorithm import BaseAlgorithm
from quantuminspire.sdk.models.job_options import JobOptions
from quantuminspire.util.api.base_backend import BaseBackend
from quantuminspire.util.authentication import Configuration, OauthDeviceSession
from quantuminspire.util.configuration import Settings


class RemoteBackend(BaseBackend):
    """Connection to remote backend.

    Create a connection with the remote backend for Quantum Inspire. This connection creates the appropriate projects,
    algorithms, files etc. The algorithm/circuit will also be run.
    """

    def __init__(self) -> None:
        super().__init__()
        settings = Settings()
        self.auth_settings = settings.default_auth_settings

        oauth_session = OauthDeviceSession(self.auth_settings)
        tokens = oauth_session.refresh()
        settings.store_tokens(settings.default_host, tokens)

        self._configuration = Configuration(host=settings.default_host, oauth_session=oauth_session)

    def run(
        self,
        program: BaseAlgorithm,
        backend_type_id: int,
        options: JobOptions = JobOptions(),
    ) -> int:
        """Execute provided algorithm/circuit."""
        return asyncio.run(self._create_flow(program, backend_type_id, job_options=options))

    async def _get_job(self, job_id: int) -> Any:
        async with ApiClient(self._configuration) as api_client:
            return await self._read_job(api_client, job_id)

    async def _get_results(self, job_id: int) -> Any:
        async with ApiClient(self._configuration) as api_client:
            job = await self._read_job(api_client, job_id)
            if job.status != JobStatus.COMPLETED:
                return None

            return await self._read_results_for_job(api_client, job)

    async def _get_final_results(self, job_id: int) -> Any:
        async with ApiClient(self._configuration) as api_client:
            job = await self._read_job(api_client, job_id)
            if job.status != JobStatus.COMPLETED:
                return None

            return await self._read_final_results_for_job(api_client, job)

    async def _get_backend_types(self) -> Any:
        async with ApiClient(self._configuration) as api_client:
            api_instance = BackendTypesApi(api_client)
            return await api_instance.read_backend_types_backend_types_get()

    def get_job(self, job_id: int) -> Any:
        """Get job for algorithm/circuit."""
        return asyncio.run(self._get_job(job_id))

    def get_results(self, job_id: int) -> Any:
        """Get results for algorithm/circuit."""
        return asyncio.run(self._get_results(job_id))

    def get_final_results(self, job_id: int) -> Any:
        """Get final results for algorithm/circuit."""
        return asyncio.run(self._get_final_results(job_id))

    def get_backend_types(self) -> Any:
        """Get backend types."""
        return asyncio.run(self._get_backend_types())

    async def _create_flow(
        self,
        program: BaseAlgorithm,
        backend_type_id: int,
        job_options: JobOptions,
    ) -> int:
        """Call the necessary methods in the correct order, with the correct parameters."""
        async with ApiClient(self._configuration) as api_client:
            project = await self._create_project(api_client, program)
            algorithm = await self._create_algorithm(api_client, program, project)
            commit = await self._create_commit(api_client, algorithm)
            file = await self._create_file(api_client, program, commit)
            batch_job = await self._create_batch_job(api_client, backend_type_id=backend_type_id)
            job: Job = await self._create_job(api_client, file, batch_job, job_options)
            await self._enqueue_batch_job(api_client, batch_job)
            return job.id  # type: ignore

    async def _create_project(self, api_client: ApiClient, program: BaseAlgorithm) -> Project:
        api_instance = ProjectsApi(api_client)
        obj = ProjectIn(
            owner_id=self.auth_settings.owner_id,
            name=program.program_name,
            description="Project created by SDK",
            starred=False,
        )
        return await api_instance.create_project_projects_post(obj)

    @staticmethod
    async def _create_algorithm(api_client: ApiClient, program: BaseAlgorithm, project: Project) -> Algorithm:
        api_instance = AlgorithmsApi(api_client)
        obj = AlgorithmIn(
            project_id=project.id,
            type=program.content_type,
            shared=ShareType.PRIVATE,
            name=program.program_name,
        )
        return await api_instance.create_algorithm_algorithms_post(obj)

    @staticmethod
    async def _create_commit(api_client: ApiClient, algorithm: Algorithm) -> Commit:
        api_instance = CommitsApi(api_client)
        obj = CommitIn(
            description="Commit created by SDK",
            algorithm_id=algorithm.id,
        )
        return await api_instance.create_commit_commits_post(obj)

    @staticmethod
    async def _get_language_for_algorithm(api_client: ApiClient, algorithm: BaseAlgorithm) -> Language:
        api_instance = LanguagesApi(api_client)
        page_languages = await api_instance.read_languages_languages_get()
        for language in page_languages.items:
            if language.name.lower() == algorithm.language_name.lower():
                return language

        raise ValueError(f"Language {algorithm.language_name} not found in API")

    @staticmethod
    async def _create_file(api_client: ApiClient, program: BaseAlgorithm, commit: Commit) -> File:
        api_instance = FilesApi(api_client)
        language = await RemoteBackend._get_language_for_algorithm(api_client, program)

        obj = FileIn(
            commit_id=commit.id,
            content=program.content,
            language_id=language.id,
            compile_stage=program.compile_stage,
            compile_properties={},
        )
        return await api_instance.create_file_files_post(obj)

    @staticmethod
    async def _create_batch_job(api_client: ApiClient, backend_type_id: int) -> BatchJob:
        api_instance = BatchJobsApi(api_client)
        obj = BatchJobIn(backend_type_id=backend_type_id)
        return await api_instance.create_batch_job_batch_jobs_post(obj)

    @staticmethod
    async def _create_job(api_client: ApiClient, file: File, batch_job: BatchJob, job_options: JobOptions) -> Job:
        api_instance = JobsApi(api_client)
        obj = JobIn(file_id=file.id, batch_job_id=batch_job.id, **job_options.model_dump())
        return await api_instance.create_job_jobs_post(obj)

    @staticmethod
    async def _enqueue_batch_job(api_client: ApiClient, batch_job: BatchJob) -> BatchJob:
        api_instance = BatchJobsApi(api_client)
        return await api_instance.enqueue_batch_job_batch_jobs_id_enqueue_patch(batch_job.id)

    @staticmethod
    async def _read_job(api_client: ApiClient, job_id: int) -> Job:
        api_instance = JobsApi(api_client)
        return await api_instance.read_job_jobs_id_get(job_id)

    @staticmethod
    async def _read_results_for_job(api_client: ApiClient, job: Job) -> List[Result]:
        api_instance = ResultsApi(api_client)
        return await api_instance.read_results_by_job_id_results_job_job_id_get(job.id)  # type: ignore

    @staticmethod
    async def _read_final_results_for_job(api_client: ApiClient, job: Job) -> List[FinalResult]:
        api_instance = FinalResultsApi(api_client)
        return await api_instance.read_final_result_by_job_id_final_results_job_job_id_get(job.id)  # type: ignore
